import os
import yaml

FILE_NAME = 'config.yaml'
BASEPATH = os.getcwd()


class ConfigError(KeyError):
    """ Raised when a requested configuration can not be found """


class ConfigBase:
    def __init__(self):
        self.basename = ''
        self.default_options = {}
        raise TypeError('Should not use ConfigBase directly')

    def __getitem__(self, item):
        """ called when access like a dictionary
                config['database']['username']
        """
        default = self.default_options.get(item)
        if isinstance(default, dict):
            return ConfigSection(self.basename + '.' + item, default)
        # Check for env var
        envvar_string = '_'.join(self.basename.split('.') + [item])\
            .strip('_').upper()
        env = os.getenv(envvar_string)
        if env is None:
            if default is None:
                raise ConfigError(
                    'No configuration found for "{}". '
                    'Add the environment variable {} or add'
                    ' the key to your config.toml file'.format(
                        '.'.join((self.basename, item)).strip('.'),
                        envvar_string
                    ))
            return default
        # env vars are always passed in as strings in docker world
        # here we will try to convert them to basic types if we can
        if env.lower() == 'true':
            env = True
        if env.lower() == 'false':
            env = False
        try:
            env = int(env)
        except ValueError:
            pass
        return env


class Config(ConfigBase):
    """ load configuration from envvar or config file
        You can get settings frhelloom environment variables or a config.yaml file
        Environment variables have a higher precedence over the config file

        Use `config['section']['subsection']['key']` to get the value
        SECTION_SUBSECTION_KEY from environment variables or
        section.subsection.key from a yaml file
        (usually written:
            [section.subsection]
            key = value
        in a toml file)
    """
    def __init__(self):
        self.default_options = None
        self.basename = ''

    def load_config(self):
        filepath = os.path.abspath(os.path.join(BASEPATH, FILE_NAME))
        with open(filepath, 'r') as f:
            config = yaml.load(f)
            return config

    def __getitem__(self, item):
        if self.default_options is None:
            self.default_options = self.load_config()
        return super().__getitem__(item)


class ConfigSection(ConfigBase):
    def __init__(self, basename, defaults):
        self.basename = basename.lstrip('.')
        self.default_options = defaults


import traceback
from typing import List, Union, Iterable
import asyncio
from http import HTTPStatus

from .client import Client
from .webtypes import Request, Response, ResponseError
from .router import Router
from waspy.transports.transportabc import TransportABC


async def response_wrapper_factory(app, handler):
    async def wrap_response_middleware(request):
        response = await handler(request)
        if not isinstance(response, Response):
            if isinstance(response, tuple):
                body = response[0]
                status = response[1]
                response = Response(status=status, body=body)
            elif isinstance(response, dict) or isinstance(response, str):
                response = Response(body=response)
            elif response is None:
                response = Response(status=HTTPStatus.NO_CONTENT)
            else:
                raise ValueError('Request handler returned an invalid type.'
                                 ' Return types should be one of '
                                 '[Response, dict, str, None, (dict, int)]')
        return response
    return wrap_response_middleware


class Application:
    def __init__(self,
                 transport: Union[TransportABC,
                                  Iterable[TransportABC]]=None,
                 *,
                 middlewares: List[callable]=None,
                 default_headers: dict=None,
                 debug: bool=False,
                 router: Router=None):
        if transport is None:
            from waspy.transports.httptransport import HTTPTransport
            transport = HTTPTransport()
        if not isinstance(transport, tuple):
            transport = (transport,)
        if middlewares is None:
            middlewares = ()
        middlewares = tuple([m for m in middlewares])
        middlewares += (response_wrapper_factory,)
        if router is None:
            router = Router()
        if default_headers is None:
            default_headers = {'Server': 'waspy'}
        self.transport = transport
        self.middlewares = middlewares
        self.default_headers = default_headers
        self.debug = debug
        self.router = router
        self.on_start = []
        self._client = None

    @property
    def client(self) -> Client:
        if not self._client:
            self._client = Client(transport=self.transport[0].get_client())
        return self._client

    def run(self):
        loop = asyncio.get_event_loop()
        # wrap handlers in middleware
        loop.run_until_complete(self._wrap_handlers())
        for t in self.transport:
            t.listen(loop=loop)

        # Call on-startup hooks
        for coro in self.on_start:
            loop.run_until_complete(coro(self))

        # todo: fork/add processes?
        for t in self.transport:
            t.start(self.handle_request, loop=loop)
        try:
            loop.run_forever()
        except KeyboardInterrupt:
            print('Shutting down')
        for t in self.transport:
            t.shutdown(loop=loop)

        # todo: Call on-shutdown hooks

        loop.close()

    async def handle_request(self, request: Request) -> Response:
        """
        coroutine: This method is called by Transport
        implementation to handle the actual request.
        It returns a webtype.Response object.
        """
        # ToDo: Send through middlewares
        # response = Response(headers={'hello': 'world'}, body={'foo': 'bar'})
        # Get handler
        try:
            handler = self.router.get_handler_for_request(request)
            response = await handler(request)

        except ResponseError as r:
            response = r.response
        except Exception as e:
            traceback.print_exc()
            response = Response(status=500)
        if not response.correlation_id:
            response.correlation_id = request.correlation_id
        # add default headers
        extra_headers = {}
        if response.data is not None:
            extra_headers['content-length'] = str(len(response.data))

        response.headers = {**self.default_headers, **extra_headers,
                            **response.headers}

        return response

    async def _wrap_handlers(self):
        handler_gen = self.router._get_and_wrap_routes()

        try:
            handler = next(handler_gen)
            while True:
                wrapped = handler
                for middleware in self.middlewares[::-1]:
                    wrapped = await middleware(self, wrapped)
                handler = handler_gen.send(wrapped)
        except StopIteration:
            pass

